/// <reference types="node" />
import { CookieJar } from 'tough-cookie';
import { Middleware } from 'throwback';
import { Agent } from 'http';
import { Request, Response, ResponseOptions } from 'servie';
import { Body } from 'servie/dist/body/node';
import { FollowRedirectsOptions } from '../common';
/**
 * Extend response with URL.
 */
export interface HttpResponseOptions extends ResponseOptions {
    body: Body;
}
/**
 * HTTP responses implement a node.js body.
 */
export declare class HttpResponse extends Response implements HttpResponseOptions {
    body: Body;
    constructor(options: HttpResponseOptions);
}
/**
 * Cookie middleware configuration.
 */
export interface CookieOptions {
    jar: CookieJar;
}
/**
 * Read cookies from the cookie jar.
 */
export declare function getCookies<T extends Request, U extends Response>(options: CookieOptions): Middleware<T, U>;
/**
 * Save response cookies into the cookie jar.
 */
export declare function saveCookies<T extends Request, U extends Response>(options: CookieOptions): Middleware<T, U>;
/**
 * Automatically support decoding zipped responses.
 */
export declare function autoUnzip<T extends Request, U extends HttpResponse>(): Middleware<T, U>;
/**
 * Set default user-agent in node.js.
 */
export declare function normalizeUserAgent<T extends Request, U extends Response>(): Middleware<T, U>;
/**
 * Node.js HTTP request options.
 */
export interface SendOptions {
    agent?: Agent;
    rejectUnauthorized?: boolean;
    ca?: string | Buffer | Array<string | Buffer>;
    cert?: string | Buffer;
    key?: string | Buffer;
    secureProtocol?: string;
}
/**
 * Function to execute HTTP request.
 */
export declare function send(options: SendOptions): (req: Request) => Promise<HttpResponse>;
/**
 * Node.js HTTP transport configuration.
 */
export interface TransportOptions extends SendOptions, FollowRedirectsOptions, SendOptions {
    jar?: CookieJar;
    unzip?: boolean;
    follow?: boolean;
}
/**
 * Create a request transport using node.js `http` libraries.
 */
export declare function transport(options?: TransportOptions): (req: Request) => Promise<HttpResponse>;
