"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const pump = require("pump");
const throwback_1 = require("throwback");
const http_1 = require("http");
const https_1 = require("https");
const stream_1 = require("stream");
const zlib_1 = require("zlib");
const servie_1 = require("servie");
const node_1 = require("servie/dist/body/node");
const url_1 = require("url");
const error_1 = require("../error");
const common_1 = require("../common");
/**
 * HTTP responses implement a node.js body.
 */
class HttpResponse extends servie_1.Response {
    constructor(options) {
        super(options);
        this.body = options.body;
    }
}
exports.HttpResponse = HttpResponse;
/**
 * Read cookies from the cookie jar.
 */
function getCookies(options) {
    return function (req, next) {
        const prevCookies = req.headers.getAll('Cookie').join('; ');
        return new Promise((resolve, reject) => {
            options.jar.getCookieString(req.url, (err, cookies) => {
                if (err)
                    return reject(err);
                if (cookies) {
                    req.headers.set('Cookie', prevCookies ? `${prevCookies}; ${cookies}` : cookies);
                }
                return resolve(next());
            });
        });
    };
}
exports.getCookies = getCookies;
/**
 * Save response cookies into the cookie jar.
 */
function saveCookies(options) {
    return function (req, next) {
        return __awaiter(this, void 0, void 0, function* () {
            const res = yield next();
            const cookies = res.headers.getAll('set-cookie');
            if (cookies.length) {
                yield Promise.all(cookies.map(function (cookie) {
                    return new Promise(function (resolve, reject) {
                        options.jar.setCookie(cookie, req.url, { ignoreError: true }, (err) => err ? reject(err) : resolve());
                    });
                }));
            }
            return res;
        });
    };
}
exports.saveCookies = saveCookies;
/**
 * Automatically support decoding zipped responses.
 */
function autoUnzip() {
    return function (req, next) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!req.headers.has('Accept-Encoding')) {
                req.headers.set('Accept-Encoding', 'gzip,deflate');
            }
            const res = yield next();
            const enc = res.headers.get('Content-Encoding');
            // Unzip body automatically when response is encoded.
            if (enc === 'deflate' || enc === 'gzip') {
                res.body = node_1.createBody(pump(res.body.stream(), zlib_1.createUnzip()), { headers: [] });
            }
            return res;
        });
    };
}
exports.autoUnzip = autoUnzip;
/**
 * Set default user-agent in node.js.
 */
function normalizeUserAgent() {
    return function (req, next) {
        if (!req.headers.has('User-Agent')) {
            req.headers.set('User-Agent', 'Popsicle (https://github.com/blakeembrey/popsicle)');
        }
        return next();
    };
}
exports.normalizeUserAgent = normalizeUserAgent;
/**
 * Function to execute HTTP request.
 */
function send(options) {
    return function (req) {
        return new Promise((resolve, reject) => {
            const { body } = req;
            const arg = url_1.parse(req.url);
            const encrypted = arg.protocol === 'https:';
            const engine = encrypted ? https_1.request : http_1.request;
            // Attach request options.
            arg.method = req.method;
            arg.headers = req.headers.asObject(false);
            arg.agent = options.agent;
            arg.rejectUnauthorized = options.rejectUnauthorized !== false;
            arg.ca = options.ca;
            arg.cert = options.cert;
            arg.key = options.key;
            arg.secureProtocol = options.secureProtocol;
            const rawRequest = engine(arg);
            const requestStream = new stream_1.PassThrough();
            // Trigger unavailable error when node.js errors before response.
            function onError(err) {
                reject(new error_1.PopsicleError('Unable to connect', 'EUNAVAILABLE', req, err));
            }
            // Track the node.js response.
            function onResponse(rawResponse) {
                const { statusCode, statusMessage } = rawResponse;
                const headers = servie_1.createHeaders(rawResponse.rawHeaders);
                const body = node_1.createBody(pump(rawResponse, new stream_1.PassThrough()));
                // Trailers are populated on "end".
                const trailer = new Promise(resolve => {
                    rawResponse.on('end', () => resolve(servie_1.createHeaders(rawResponse.rawTrailers)));
                });
                // Replace request error listener behaviour.
                rawRequest.removeListener('error', onError);
                rawRequest.on('error', err => req.events.emit('error', err));
                const { address: localAddress, port: localPort } = rawRequest.connection.address();
                const { address: remoteAddress, port: remotePort } = rawResponse.connection.address();
                const res = new HttpResponse({ statusCode, statusMessage, headers, trailer, body });
                // Update request connection.
                req.connection = { localAddress, localPort, remoteAddress, remotePort, encrypted };
                // https://github.com/serviejs/servie#implementers
                res.started = true;
                req.events.emit('response', res);
                // Track response progress.
                rawResponse.on('data', (chunk) => {
                    res.bytesTransferred += chunk.length;
                });
                rawResponse.on('end', () => res.finished = true);
                rawResponse.on('close', () => req.closed = true);
                return resolve(res);
            }
            // Track request upload progress.
            requestStream.on('data', (chunk) => {
                req.bytesTransferred += chunk.length;
            });
            // Listen for connection errors.
            rawRequest.once('error', onError);
            rawRequest.once('response', onResponse);
            // https://github.com/serviejs/servie#implementers
            req.started = true;
            req.events.on('abort', () => rawRequest.abort());
            // Pump request body into HTTP request object.
            pump(requestStream, rawRequest, () => req.finished = true);
            // Pipe the body to the stream.
            if (body instanceof node_1.Body) {
                if (body.buffered) {
                    body.buffer().then(x => requestStream.end(x), reject);
                }
                else {
                    pump(body.stream(), requestStream);
                }
            }
            else {
                body.arrayBuffer().then(x => requestStream.end(Buffer.from(x)), reject);
            }
        });
    };
}
exports.send = send;
/**
 * Create a request transport using node.js `http` libraries.
 */
function transport(options = {}) {
    const fns = [common_1.normalizeRequest(), normalizeUserAgent()];
    const { jar, unzip = true, follow = true } = options;
    if (unzip)
        fns.push(autoUnzip());
    if (follow)
        fns.push(common_1.followRedirects(options));
    if (jar)
        fns.push(getCookies({ jar }), saveCookies({ jar }));
    const done = send(options);
    const middlware = throwback_1.compose(fns);
    return (req) => middlware(req, done);
}
exports.transport = transport;
//# sourceMappingURL=http.js.map